/*****************************************************************************

  Licensed to Accellera Systems Initiative Inc. (Accellera) under one or
  more contributor license agreements.  See the NOTICE file distributed
  with this work for additional information regarding copyright ownership.
  Accellera licenses this file to you under the Apache License, Version 2.0
  (the "License"); you may not use this file except in compliance with the
  License.  You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
  implied.  See the License for the specific language governing
  permissions and limitations under the License.

 *****************************************************************************/

// low_pass.cpp  - regression test for different kinds of equation description

#include <systemc-ams>
#include <string>

#include "test_utilities.h"

// lowpass implementation as non-conservative behavioral model with sca_ltf_nd
SCA_TDF_MODULE (low_nd)
{
  // ports
  sca_tdf::sca_in<double> inp;
  sca_tdf::sca_out<double> outp;

  struct params // parameters
  {
    double fc;

    params() // default for parameter
    {
      fc = 1.0e3;
    }
  };

  void ac_processing(); // ac-implementation
  void initialize();    // initialization
  void processing();    // time domain processing method

  // constructor
  low_nd( sc_core::sc_module_name nm, params pa = params())
  : inp("inp"), outp("outp"), ltf1("ltf1"), p(pa)
  {}

  // definition of local variables
private:
  sca_tdf::sca_ltf_nd ltf1;
  sca_util::sca_vector<double> A, B, S;
  params p;
};

// implementation of methods

void low_nd::initialize()
{
  B(0) = 1.0;
  A(0) = 1.0;
  A(1) = 1.0 / (2.0 * M_PI * p.fc);
}

// frequency domain implementation
void low_nd::ac_processing()
{
  sca_ac_analysis::sca_ac(outp) = sca_ac_analysis::sca_ac_ltf_nd(B, A, sca_ac_analysis::sca_ac(inp));
}

// time domain implementation
void low_nd::processing()
{
  outp = ltf1(B, A, S, inp);
}

// lowpass implementation as non-conservative behavioral model with sca_ltf_zp
SCA_TDF_MODULE (low_zp)
{
  // ports
  sca_tdf::sca_in<double> inp;
  sca_tdf::sca_out<double> outp;

  struct params // parameters
  {
    double fc;
    double K;
    params() // default for parameters
    {
      fc = 1.0e3;
      K = 1.0;
    }
  };

  void ac_processing(); // ac-implementation
  void initialize();    // initialization
  void processing();    // time domain processing method

  // constructor
  low_zp( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), ltf1("ltf1"), p(pa)
  {}

  // definition of local variables
private:
  sca_tdf::sca_ltf_zp ltf1;
  sca_util::sca_vector<sca_util::sca_complex> Z, P;
  sca_util::sca_vector<double> S;
  params p;
};

// implementation of methods

// vector initialization
void low_zp::initialize()
{
  // H(s) = K / (s-P(0))
  P(0) = -2.0 * M_PI * p.fc;
  p.K  =  2.0 * M_PI * p.fc;
}

// frequency domain
void low_zp::ac_processing()
{
  sca_ac_analysis::sca_ac(outp) = sca_ac_analysis::sca_ac_ltf_zp(Z, P, sca_ac_analysis::sca_ac(inp), p.K);
}

// time domain
void low_zp::processing()
{
  outp = ltf1(Z, P, S, inp, p.K);
}

// lowpass implementation as hierarchical module with a non-conservative inport
// and a conservative (electrical) outport and implemented as conservative RC-network

SC_MODULE(low_rc)
{
  // port and terminal declaration
  sca_tdf::sca_in<double> inp;
  sca_eln::sca_terminal outp;

  // component declaration
  sca_eln::sca_tdf_vsource* vin;
  sca_eln::sca_r* r1;
  sca_eln::sca_c* c1;

  // node declaration
  sca_eln::sca_node n_in;
  sca_eln::sca_node_ref gnd; // reference node

  struct params // parameters
  {
    double r_val;
    double c_val;

    params() // default for parameters
    {
      double fc = 1.0e3; // default for 1kHz cut off freq.

      r_val = 1.0e3;
      c_val = 1.0 / (2.0 * M_PI * fc * r_val);
    }
  };

  low_rc( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), n_in("n_in"), gnd("gnd"), p(pa)
  {
    architecture(); // generate netlist
  }

  ~low_rc();

private:
  params p;
  void architecture(); // method for netlist
};

// netlist implementation
void low_rc::architecture()
{
  vin = new sca_eln::sca_tdf_vsource("vin");
  vin->inp(inp);
  vin->p(n_in);
  vin->n(gnd);

  r1 = new sca_eln::sca_r("r1");
  r1->value = p.r_val; // the value can be assigned via the member...
  r1->n(n_in);
  r1->p(outp);

  c1 = new sca_eln::sca_c("c1", p.c_val); // or via the constructor
  c1->p(outp);
  c1->n(gnd);
}

low_rc::~low_rc()
{
  delete vin, r1, c1;
}

// lowpass implementation as hierarchical module with a LSF

SC_MODULE(low_lsf_nd)
{
  sca_tdf::sca_in<double> inp;
  sca_lsf::sca_out outp;

  struct params // parameters
  {
    double fc;

    params() // default for parameters
    {
      fc = 1.0e3; // default for 1kHz cut off freq.
    }
  };

  //component declaration
  sca_lsf::sca_tdf_source* tdf2lsf;
  sca_lsf::sca_ltf_nd* ltf_nd;

  //node declaration
  sca_lsf::sca_signal s1;

  low_lsf_nd( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), s1("s1"), p(pa)
  {
    architecture(); // method for netlist
  }

  ~low_lsf_nd();

private:
  params p;
  void architecture(); // method for netlist
};

// netlist implementation
void low_lsf_nd::architecture()
{
  tdf2lsf = new sca_lsf::sca_tdf_source("tdf2lsf");
  tdf2lsf->inp(inp);
  tdf2lsf->y(s1);

  ltf_nd = new sca_lsf::sca_ltf_nd("ltf_nd");
  ltf_nd->num = sca_util::sca_create_vector(1.0);
  ltf_nd->den = sca_util::sca_create_vector(1.0, 1.0 / (2.0 * M_PI * p.fc));
  ltf_nd->x(s1);
  ltf_nd->y(outp);
}

low_lsf_nd::~low_lsf_nd()
{
  delete tdf2lsf, ltf_nd;
}

// lowpass implementation as hierarchical module with a LSF wit zero-pole

SC_MODULE(low_lsf_zp)
{
  sca_tdf::sca_in<double> inp;
  sca_lsf::sca_out outp;

  struct params // parameter
  {
    double fc;

    params() // default for parameter
    {
      fc = 1.0e3; // default for 1kHz cut off freq.
    }
  };

  //component declaration
  sca_lsf::sca_tdf_source* tdf2lsf;
  sca_lsf::sca_ltf_zp* ltf_zp;

  //node declaration
  sca_lsf::sca_signal s1;

  low_lsf_zp( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), s1("s1"), p(pa)
  {
    architecture();
  }

  ~low_lsf_zp();

private:
  params p;
  void architecture(); // method for netlist
};

//netlist implementation
void low_lsf_zp::architecture()
{
  tdf2lsf = new sca_lsf::sca_tdf_source("tdf2lsf");
  tdf2lsf->inp(inp);
  tdf2lsf->y(s1);

  ltf_zp = new sca_lsf::sca_ltf_zp("ltf_zp");
  ltf_zp->poles = sca_util::sca_create_vector<sca_util::sca_complex> (-2.0 * M_PI * p.fc);
  ltf_zp->k = 2.0 * M_PI * p.fc;
  ltf_zp->x(s1);
  ltf_zp->y(outp);
}

low_lsf_zp::~low_lsf_zp()
{
  delete tdf2lsf, ltf_zp;
}

// lowpass implementation as non-conservative behavioral model with sca_ss

SCA_TDF_MODULE (lowpass_ss)
{
  // ports
  sca_tdf::sca_in<double>  inp;
  sca_tdf::sca_out<double> outp;

  struct params // parameter
  {
    double fc;

    params() // default for parameter
    {
      fc = 1.0e3;
    }
  };

  void ac_processing();  // ac-implementation
  void initialize();     // initialization
  void processing();     // time domain processing method
 
  // constructor
  lowpass_ss( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), ss1("ss1"), p(pa)
  {}

  // definition of local variables
private:
  sca_tdf::sca_ss  ss1;
  sca_util::sca_vector<double> S;
  sca_util::sca_matrix<double> A, B, C, D;
  params p;
};

//implementation of methods

void lowpass_ss::initialize()
{
  double r_val = 1.0e3;
  double c_val = 1.0 / (2.0 * M_PI * p.fc * r_val);

  // dQ=i=1/r * (vin-vout)  ; vout = 1/c *Q
  // dQ   = -1/r/c * Q    + 1/r  * vin
  // vout =   1/c  * Q    +  0.0 * vin
  A(0, 0) = -1.0 / c_val / r_val;
  B(0, 0) = 1.0 / r_val;

  C(0, 0) = 1.0 / c_val;
  D(0, 0) = 0.0;
}

// frequency domain implementation
void lowpass_ss::ac_processing()
{
  sca_util::sca_vector<sca_util::sca_complex> tmp_in(1), tmp_out;
  tmp_in(0) = sca_ac_analysis::sca_ac(inp);
  tmp_out = sca_ac_analysis::sca_ac_ss(A, B, C, D, tmp_in);
  sca_ac_analysis::sca_ac(outp) = tmp_out(0);
}

// time domain implementation
void lowpass_ss::processing()
{
  sca_util::sca_vector<double> tmp_in, tmp_out;
  tmp_in(0) = inp.read();
  tmp_out = ss1(A, B, C, D, S, tmp_in);
  outp = tmp_out(0);
}

// lowpass implementation as hierarchical module with a non-conservative inport
// and a lsf network with sca_lsf::sca_ss module

SC_MODULE(lowpass_lsf_ss)
{
  // port declaration
  sca_tdf::sca_in<double> inp;
  sca_lsf::sca_out        outp;

  // component declaration
  sca_lsf::sca_tdf_source* value_in;
  sca_lsf::sca_ss*         i_ss;

  // signal declaration
  sca_lsf::sca_signal s_in;

  struct params // parameters
  {
    double fc;

    params() // default for parameters
    {
      fc = 1.0e3; // default for 1kHz cut off freq.
    }
  };
  // constructor
  lowpass_lsf_ss( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), s_in("s_in"), p(pa)
  {
    architecture(); // generate netlist
  }

  ~lowpass_lsf_ss();

private:
  params p;
  void architecture(); // method for netlist
};


// netlist implementation
void lowpass_lsf_ss::architecture()
{
  double r_val = 1.0e3;
  double c_val = 1.0 / (2.0 * M_PI * p.fc * r_val);
  sca_util::sca_matrix<double> A, B, C, D;

  A(0, 0) = -1.0/c_val/r_val;
  B(0, 0) = 1.0/r_val;

  C(0, 0) = 1.0/c_val;
  D(0, 0) = 0.0;


  value_in = new sca_lsf::sca_tdf_source("value_in");
  value_in->inp(inp);
  value_in->y(s_in);

  i_ss = new sca_lsf::sca_ss("i_ss", A, B, C, D);
  i_ss->x(s_in);
  i_ss->y(outp);
}

lowpass_lsf_ss::~lowpass_lsf_ss()
{
  delete value_in, i_ss;
}

// lowpass implementation as hierarchical module with a non-conservative inport
// and a lsf network

SC_MODULE(lowpass_lsf)
{
  // port declaration
  sca_tdf::sca_in<double> inp;
  sca_lsf::sca_out        outp;

  // component declaration
  sca_lsf::sca_tdf_source* value_in;
  sca_lsf::sca_dot*        dt;
  sca_lsf::sca_sub*        sub;
  sca_lsf::sca_gain*       gain;

  // signal declaration
  sca_lsf::sca_signal s_in, s_1, s_2;

  struct params // parameters
  {
    double fc;

    params() // default for parameters
    {
      fc = 1.0e3; // default for 1kHz cut off freq.
    }
  };

  lowpass_lsf( sc_core::sc_module_name nm, params pa = params() )
  : inp("inp"), outp("outp"), s_in("s_in"), s_1("s_1"), s_2("s_2"), p(pa)
  {
    architecture(); // generate netlist
  }

  ~lowpass_lsf();

private:
  params p;
  void architecture(); // method for netlist
};

// netlist implementation
void lowpass_lsf::architecture()
{
  value_in = new sca_lsf::sca_tdf_source("value_in");
  value_in->inp(inp);
  value_in->y(s_in);

  sub = new sca_lsf::sca_sub("sub");
  sub->x1(s_in);
  sub->x2(s_1);
  sub->y(outp);

  gain = new sca_lsf::sca_gain("gain");
  gain->x(s_2);
  gain->y(s_1);
  gain->k = 1.0 / 2.0 / M_PI / p.fc;

  dt = new sca_lsf::sca_dot("dt");
  dt->x(outp);
  dt->y(s_2);
}

lowpass_lsf::~lowpass_lsf()
{
  delete value_in, sub, gain, dt;
}

// non-conservative sine wave source

SCA_TDF_MODULE(sin_src)
{
  sca_tdf::sca_out<double> outp;

  struct params
  {
    double freq;
    double ampl;
    sc_core::sc_time sample_time;

    params()
    {
      ampl = 1.0;
      freq = 1e3;
      sample_time = sc_core::sc_time(10.0, sc_core::SC_US);
    }
  };

  void set_attributes()
  {
    outp.set_timestep(p.sample_time);
  }

  void ac_processing();
  void processing();

  sin_src( sc_core::sc_module_name nm, params pa = params() )
  : outp("outp"), p(pa)
  {
  }

private:
  params p;
};

// frequency domain implementation
void sin_src::ac_processing()
{
  // we use for ac-domain for all frequencies the same amplitude and a phase=0
  sca_ac_analysis::sca_ac(outp) = p.ampl;
}

// time domain implementation
void sin_src::processing()
{
  outp = p.ampl * std::sin(2.0 * M_PI * p.freq * sc_core::sc_time_stamp().to_seconds());
}

// testbench implementation
void check_td_results(std::string fname);
void check_ac_results(std::string fname);

int sc_main(int argn, char* argc[])
{
  TEST_LABEL_START;

  // define non-conservative signals
  sca_tdf::sca_signal<double> s_in, s_out_nd, s_out_zp, s_out_tdf_ss;

  // define electrical node
  sca_eln::sca_node n_out_rc;

  // lsf signal
  sca_lsf::sca_signal s_lsf_nd_out, s_lsf_zp_out, s_out_lsf_ss, s_out_lsf_hier;

  // instantiate and connect components

  sin_src* i_src = new sin_src("i_src");
  i_src->outp(s_in);

  ////////////////////////////////////////////////

  low_nd* i_lp_nd = new low_nd("i_lp_nd");
  i_lp_nd->inp(s_in);
  i_lp_nd->outp(s_out_nd);

  ////////////////////////////////////////////////

  low_zp* i_lp_zp = new low_zp("i_lp_zp");
  i_lp_zp->inp(s_in);
  i_lp_zp->outp(s_out_zp);

  ////////////////////////////////////////////////

  low_rc* i_lp_rc = new low_rc("i_lp_rc");
  i_lp_rc->inp(s_in);
  i_lp_rc->outp(n_out_rc);

  ////////////////////////////////////////////////

  low_lsf_nd* i_lp_lsf_nd = new low_lsf_nd("i_lp_lsf_nd");
  i_lp_lsf_nd->inp(s_in);
  i_lp_lsf_nd->outp(s_lsf_nd_out);

  ////////////////////////////////////////////////

  low_lsf_zp* i_lp_lsf_zp = new low_lsf_zp("i_lp_lsf_zp");
  i_lp_lsf_zp->inp(s_in);
  i_lp_lsf_zp->outp(s_lsf_zp_out);

  ////////////////////////////////////////////////

  lowpass_ss* i_lp_tdf_ss = new lowpass_ss("i_lp_tdf_ss");
  i_lp_tdf_ss->inp(s_in);
  i_lp_tdf_ss->outp(s_out_tdf_ss);

  ////////////////////////////////////////////////

  lowpass_lsf_ss* i_lp_lsf_ss = new lowpass_lsf_ss("i_lp_lsf_ss");
  i_lp_lsf_ss->inp(s_in);
  i_lp_lsf_ss->outp(s_out_lsf_ss);

  ////////////////////////////////////////////////

  lowpass_lsf* i_lp_lsf_hier = new lowpass_lsf("i_lp_lsf_hier");
  i_lp_lsf_hier->inp(s_in);
  i_lp_lsf_hier->outp(s_out_lsf_hier);

  ////////////////////////////////////////////////

  sca_util::sca_trace_file* tf = sca_util::sca_create_tabular_trace_file("low_pass_td.dat");
  sca_util::sca_trace(tf, s_in, "s_in");
  sca_util::sca_trace(tf, s_out_nd, "s_out_nd");
  sca_util::sca_trace(tf, s_out_zp, "s_out_zp");
  sca_util::sca_trace(tf, n_out_rc, "n_out_rc"); //voltage at node
  sca_util::sca_trace(tf, s_lsf_nd_out, "s_lsf_nd_out");
  sca_util::sca_trace(tf, s_lsf_zp_out, "s_lsf_zp_out");
  sca_util::sca_trace(tf, s_out_tdf_ss, "s_out_tdf_ss");
  sca_util::sca_trace(tf, s_out_lsf_ss, "s_out_lsf_ss");
  sca_util::sca_trace(tf, s_out_lsf_hier, "s_out_lsf_hier");
  sca_util::sca_trace(tf, i_lp_rc->n_in, "i_lp_rc.n_in"); // voltage of internal node
  sca_util::sca_trace(tf, i_lp_rc->r1, "i_lp_rc.r1"); // current through r1
  sca_util::sca_trace(tf, i_lp_rc->c1, "i_lp_rc.c1"); // current through c1

  ////////////////////////////////////////////////

  sc_core::sc_start(5.0, sc_core::SC_MS); // start time domain simulation for 5ms

  std::cout << sc_core::sc_time_stamp() << " time domain simulation finished" << std::endl
            << std::endl;

  ////////////////////////////////////////////////

  // we store the ac result in another file
  tf->reopen("low_pass_ac.dat");
  // we store the ac result as magnitude in dB and phase in degree
  // instead the default complex numbers
  tf->set_mode(sca_util::sca_ac_format(sca_util::SCA_AC_DB_DEG));

  // start ac-simulation from 100Hz to 10kHz with 200 points logarithmically spaced
  sca_ac_analysis::sca_ac_start(10.0, 100.0e3, 400, sca_ac_analysis::SCA_LOG);

  sca_util::sca_close_tabular_trace_file(tf);
  
  std::cout << "AC-simulation finished" << std::endl;

  /////////////////////////////////////////////////////////////////////////
  //// Result check
  ////////////////////////////////////////////////////////////////////////

  std::cout << std::endl << "---------------------------------------------------------";
  std::cout << std::endl << "Check results of time domain simulation ..." << std::endl;

  test_util::check_results("low_pass_td", 12, "lowpass");

  std::cout << std::endl << "---------------------------------------------------------";
  std::cout << std::endl << "Check results of ac domain simulation ..." << std::endl;

  test_util::check_results("low_pass_ac", 20, "lowpass", 1e-13);

  delete i_src, i_lp_nd, i_lp_zp, i_lp_rc, i_lp_lsf_nd, i_lp_lsf_zp;
  delete i_lp_tdf_ss, i_lp_lsf_ss, i_lp_lsf_hier;

  TEST_LABEL_END;

  return 0;
}
